#!/usr/bin/env node

/**
 * windows-window-controller
 * controll window by calling user32.ShowWindow from node via node-ffi
 *
 * @version 3.2.1
 * @author waiting
 * @license MIT
 * @link https://github.com/waitingsong/node-windows-window-controller#readme
 */

'use strict';

var yargs = require('yargs');
var ffi = require('ffi');
var ref = require('ref');
var win32Api = require('win32-api');

const UC = win32Api.User32.constants;

const filterWinRulesDefaults = {
    titleExits: null,
    includeStyle: null,
    excludeStyle: null,
    includeExStyle: null,
    excludeExStyle: null,
};
const showFilterRulesDefaults = {
    titleExits: true,
    includeStyle: UC.WS_SYSMENU,
    excludeStyle: UC.WS_CHILD,
    includeExStyle: null,
    excludeExStyle: UC.WS_EX_TOOLWINDOW,
};

const knl32 = win32Api.K.load();

function get_last_err_msg() {
    // https://msdn.microsoft.com/en-us/library/windows/desktop/ms681382(v=vs.85).aspx
    const errcode = knl32.GetLastError();
    let errMsg = '';

    /* istanbul ignore if  */
    if (errcode) {
        const len = 255;
        const buf = Buffer.alloc(len);
        // tslint:disable-next-line:no-bitwise
        const p = 0x00001000 | 0x00000200; // FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS
        const langid = 0x0409; // 0x0409: US, 0x0000: Neutral locale language
        const msglen = knl32.FormatMessageW(p, null, errcode, langid, buf, len, null);

        errMsg = msglen
            ? errcode + ': ' + ref.reinterpretUntilZeros(buf, 2).toString('ucs2')
            : `${errcode}: unknown error meeage`;
    }
    return errMsg
}

const user32 = win32Api.U.load();

const taskConfig = {
    tno: 0,
    task: new Map(),
};
const enumWindowsProc = ffi.Callback(win32Api.DTypes.BOOL, [win32Api.DTypes.HWND, win32Api.DTypes.LPARAM], (hWnd, lParam) => {
    const task = taskConfig.task.get(lParam);

    /* istanbul ignore if  */
    if (!task) {
        console.error('enumWindowsProc() task not exists');
        return true
    }
    switch (task.matchType) {
        case 'pid': {
            const buf = ref.alloc(win32Api.DTypes.HINSTANCE);
            const tid = user32.GetWindowThreadProcessId(hWnd, buf);

            /* istanbul ignore else */
            if (tid) {
                const pid = buf.readUInt32LE(0);

                if (pid && pid === task.matchValue) {
                    // console.log('got by pid:', pid, 'hWnd:' + ref.address(hWnd), ref.address(hWnd).toString(16));
                    task.hwndSet.add(hWnd);
                }
            }
            else {
                const lastErrMsg = get_last_err_msg();

                if (lastErrMsg) {
                    task.errMsg += ',' + lastErrMsg;
                }
            }
            break
        }
        case 'title':
            /* istanbul ignore else */
            if (task.matchValue) {
                const buf = Buffer.alloc(254);
                const len = user32.GetWindowTextW(hWnd, buf, 254);

                if (len) {
                    const name = buf.toString('ucs2');

                    // const visible = user32.IsWindowVisible(hWnd);
                    if (name.indexOf(task.matchValue) !== -1) {
                        const buf2 = Buffer.alloc(4);

                        user32.GetWindowThreadProcessId(hWnd, buf2);
                        task.pidSet.add(buf2.readUIntLE(0, 4));
                        task.hwndSet.add(hWnd);
                    }
                }
                else {
                    const lastErrMsg = get_last_err_msg();

                    if (lastErrMsg) {
                        task.errMsg += ',' + lastErrMsg;
                    }
                }
            }
            else { // all
                task.hwndSet.add(hWnd);
            }
            break
        case 'hwnd':
            if (+task.matchValue === ref.address(hWnd)) {
                task.hwndSet.add(hWnd);
            }
            break
        /* istanbul ignore next */
        default:
            return true
    }
    return true
});
function validate_cmdshow(nCmdShow) {
    const res = (typeof nCmdShow !== 'number' || nCmdShow < 0) ? false : true;

    if (!res) {
        // tslint:disable-next-line
        console.error('hWnd value invalid: See: https://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx');
    }
    return res
}
function show_hide_one(hWnd, nCmdShow) {
    return new Promise((resolve, reject) => {
        nCmdShow = +nCmdShow;
        // console.log('show_hide_one: ' + ref.address(hWnd) + ', cmd:' + nCmdShow);
        if (Number.isNaN(nCmdShow) || nCmdShow < 0) {
            return reject('show_hide_one() params nCmdShow invalid: ' + nCmdShow)
        }
        if (nCmdShow === 0) {
            if (!user32.IsWindowVisible(hWnd)) { // skip if invisible
                return resolve()
            }
        }
        user32.ShowWindow(hWnd, nCmdShow);
        resolve(hWnd);
    })
}
function task_get_hwnds(task) {
    let t;

    return Promise.race([
        new Promise(resolve => {
            t = setTimeout(() => {
                console.info('task_get_hwnds timeout failed');
                resolve();
            }, 30000); // @HARDCOD
        }),
        _get_hwnds(task),
    ]).then((res) => {
        clearTimeout(t);
        task && taskConfig.task.delete(task.tno);
        if (res && Array.isArray(res) && res.length) {
            return res
        }
    }).catch(err => {
        clearTimeout(t);
        task && taskConfig.task.delete(task.tno);
        console.error(err);
    })
}
function _get_hwnds(task) {
    if (task.matchType === null) {
        if (typeof task.matchValue === 'string') {
            task.matchType = 'title';
        }
        else if (typeof task.matchValue === 'number') { // hwnd must be specified explicitly
            task.matchType = 'pid';
        }
    }
    switch (task.matchType) {
        case 'pid':
            if (!Number.isSafeInteger(task.matchValue)) {
                return Promise.resolve()
            }
            break
        case 'title':
            // void
            break
        case 'hwnd':
            // void
            break
        /* istanbul ignore next */
        default:
            throw new Error('task.matchType value invalid')
    }
    return get_task_hwnd(task)
        .then(res => {
        return res
    })
        .catch(err => {
        console.error(err);
        return
    })
}
function get_task_hwnd(task) {
    return new Promise((resolve, reject) => {
        if (!task) {
            resolve([]);
            return
        }
        user32.EnumWindows.async(enumWindowsProc, task.tno, (err) => {
            // console.log('get_task_hwnd size:', task.hwndSet.size);
            if (task.errMsg) {
                console.error(task.errMsg);
            }
            resolve(Array.from(task.hwndSet) || []);
        });
    })
}
// get hWnd of main top-level window
function filter_hwnd(arr, opts) {
    opts = Object.assign({}, filterWinRulesDefaults, opts);
    const ids = new Set();

    // console.log('filter_main_hwnd:', arr.length, ref.address(arr[0]));
    for (const hWnd of arr) {
        /* istanbul ignore if  */
        if (ids.has(hWnd)) {
            continue
        }
        /* istanbul ignore else  */
        if (_filter_hwnd(hWnd, opts)) {
            ids.add(hWnd);
        }
    }
    // console.log('filter_main_hwnd return size', ids.size);
    return [...ids.keys()]
}
/**
 * filter none top-level window
 * NOTE: not accurate, IME, MESSAGE windows not be filtered out
 */
function _filter_hwnd(hWnd, opts) {
    let p = user32.GetParent(hWnd);

    // const addr = ref.address(hWnd);
    // console.log('addr: ', addr + ':' + addr.toString(16));
    if (!ref.isNull(p)) {
        return false
    }
    p = user32.GetWindow(hWnd, 4); // GW_OWNER==4
    /* istanbul ignore else  */
    if (!ref.isNull(p)) {
        return false
    }
    /* istanbul ignore else  */
    if (!validate_rule_title(hWnd, opts)) {
        return false
    }
    /* istanbul ignore else  */
    if (!validate_rule_style(hWnd, opts)) {
        return false
    }
    /* istanbul ignore else  */
    if (!validate_rule_exstyle(hWnd, opts)) {
        return false
    }
    return true
}
function validate_rule_title(hWnd, opts) {
    const buf = Buffer.alloc(100);

    user32.GetWindowTextW(hWnd, buf, 50);
    let title = ref.reinterpretUntilZeros(buf, 2).toString('ucs2');

    title && (title = title.trim());
    if (opts.titleExits === true) {
        return title ? true : false
    }
    else if (opts.titleExits === false) {
        return title ? false : true
    }
    return true // 'ignore'
}
function validate_rule_style(hWnd, opts) {
    const GWL_STYLE = -16;
    const dwStyle = win32Api.Config._WIN64
        ? user32.GetWindowLongPtrW(hWnd, GWL_STYLE)
        : user32.GetWindowLongW(hWnd, GWL_STYLE);

    // console.log('style:', dwStyle);
    // if (dwStyle <= 0) {
    //     return false;
    // }
    /* istanbul ignore else  */
    if (opts.includeStyle && Number.isSafeInteger(opts.includeStyle)) {
        // tslint:disable-next-line:no-bitwise
        if ((dwStyle | opts.includeStyle) !== dwStyle) {
            return false
        }
    }
    /* istanbul ignore else  */
    if (opts.excludeStyle && Number.isSafeInteger(opts.excludeStyle)) {
        // tslint:disable-next-line:no-bitwise
        if ((dwStyle | opts.excludeStyle) === dwStyle) {
            return false
        }
    }
    return true
}
function validate_rule_exstyle(hWnd, opts) {
    const GWL_EXSTYLE = -20;
    const dwExStyle = win32Api.Config._WIN64
        ? user32.GetWindowLongPtrW(hWnd, GWL_EXSTYLE)
        : user32.GetWindowLongW(hWnd, GWL_EXSTYLE);

    // console.log('dwExstyle:' + dwExStyle);
    // if (dwExStyle <= 0) {
    //     return false;
    // }
    // if (dwExStyle && ((dwExStyle | WS_EX_TOOLWINDOW) === dwExStyle)) {
    //     return false;
    // }
    /* istanbul ignore else  */
    if (opts.includeExStyle && Number.isSafeInteger(opts.includeExStyle)) {
        // tslint:disable-next-line:no-bitwise
        if ((dwExStyle | opts.includeExStyle) !== dwExStyle) {
            return false
        }
    }
    /* istanbul ignore else  */
    if (opts.excludeExStyle && Number.isSafeInteger(opts.excludeExStyle)) {
        /* istanbul ignore else  */
        // tslint:disable-next-line:no-bitwise
        if ((dwExStyle | opts.excludeExStyle) === dwExStyle) {
            return false
        }
    }
    return true
}
function create_task() {
    const tno = ++taskConfig.tno;
    const task = {
        tno,
        matchType: null,
        matchValue: '',
        hwndSet: new Set(),
        pidSet: new Set(),
        errMsg: '',
    };

    taskConfig.task.set(tno, task);
    return task
}

const user32$1 = win32Api.U.load();

// hide the window(s)
function hide(options) {
    const opts = Object.assign({}, filterWinRulesDefaults, options);

    opts.nCmdShow = 0; /* SW_HIDE */
    return proxy(opts)
}
// https://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx
function proxy(opts) {
    const execRet = init_execret();

    if (typeof opts.nCmdShow === 'undefined' || !validate_cmdshow(opts.nCmdShow)) {
        execRet.err = 1;
        execRet.msg = 'value of nCmdShow invalid';
        return Promise.resolve(execRet)
    }
    return get_hwnds(opts).then(hWnds => {
        if (hWnds && hWnds.length) {
            for (const hWnd of hWnds) {
                if (hWnd && !ref.isNull(hWnd)) {
                    // console.log('hWnd addr:', ref.address(hWnd));
                    show_hide_one(hWnd, opts.nCmdShow)
                        .then(hWndRet => hWndRet && !ref.isNull(hWndRet) && execRet.hwnds.push(ref.address(hWndRet)))
                        .catch((err) => {
                        execRet.err = 1;
                        execRet.msg += '\n ' + err;
                    });
                }
            }
        }
    })
        .catch(err => {
        execRet.err = 1;
        execRet.msg += err.toString();
    })
        .then(() => execRet)
}
/**
 * retrieve hWnds by matchValue matched by pid|title
 */
function get_hwnds(opts) {
    // console.log('opts:', opts);
    const task = create_task();

    task.matchType = opts.matchType;
    task.matchValue = opts.matchValue;
    return task_get_hwnds(task)
        .then((arr) => {
        if (arr && Array.isArray(arr) && arr.length) {
            return filter_hwnd(arr, opts)
        }
    })
}
function init_execret() {
    return {
        err: 0,
        msg: '',
        pids: [],
        hwnds: [],
    }
}
function parse_cli_opts(argv) {
    const opts = Object.assign({}, showFilterRulesDefaults);

    if (argv.pid) {
        opts.matchType = 'pid';
        opts.matchValue = +argv.pid;
    }
    else if (argv.title) {
        opts.matchType = 'title';
        opts.matchValue = argv.title;
    }
    else if (argv.hwnd) {
        opts.matchType = 'hwnd';
        opts.matchValue = argv.hwnd;
    }
    else {
        return
    }
    if (typeof argv.status !== 'undefined') {
        opts.nCmdShow = +argv.status;
        console.assert(typeof opts.nCmdShow === 'number', 'nCmdShow must typeof number');
    }
    return opts
}

/**
 * hide window(s)
 * arguments:
 * --pid={integer}
 * --title={string}
 * --hwnd={integer}
 */
const opts = parse_cli_opts(yargs.argv);

if (!opts) {
    console.error('argv empty. arguments: --pid|title|hwnd');
    console.info('status value See: https://msdn.microsoft.com/en-us/library/windows/desktop/ms633548(v=vs.85).aspx');
    process.exit(1);
}
else {
    hide(opts).then((execRet) => {
        console.info('process ret:', execRet);
        process.exit(execRet.err);
    });
}
