#!/usr/bin/env node

/**
 * windows-window-controller
 * controll window by calling user32.ShowWindow from node via node-ffi
 *
 * @version 3.2.1
 * @author waiting
 * @license MIT
 * @link https://github.com/waitingsong/node-windows-window-controller#readme
 */

'use strict';

var yargs = require('yargs');
var ffi = require('ffi');
var ref = require('ref');
var win32Api = require('win32-api');

const UC = win32Api.User32.constants;
const showFilterRulesDefaults = {
    titleExits: true,
    includeStyle: UC.WS_SYSMENU,
    excludeStyle: UC.WS_CHILD,
    includeExStyle: null,
    excludeExStyle: UC.WS_EX_TOOLWINDOW,
};

const knl32 = win32Api.K.load();

function get_last_err_msg() {
    // https://msdn.microsoft.com/en-us/library/windows/desktop/ms681382(v=vs.85).aspx
    const errcode = knl32.GetLastError();
    let errMsg = '';

    /* istanbul ignore if  */
    if (errcode) {
        const len = 255;
        const buf = Buffer.alloc(len);
        // tslint:disable-next-line:no-bitwise
        const p = 0x00001000 | 0x00000200; // FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS
        const langid = 0x0409; // 0x0409: US, 0x0000: Neutral locale language
        const msglen = knl32.FormatMessageW(p, null, errcode, langid, buf, len, null);

        errMsg = msglen
            ? errcode + ': ' + ref.reinterpretUntilZeros(buf, 2).toString('ucs2')
            : `${errcode}: unknown error meeage`;
    }
    return errMsg
}

const user32 = win32Api.U.load();

const taskConfig = {
    tno: 0,
    task: new Map(),
};
const enumWindowsProc = ffi.Callback(win32Api.DTypes.BOOL, [win32Api.DTypes.HWND, win32Api.DTypes.LPARAM], (hWnd, lParam) => {
    const task = taskConfig.task.get(lParam);

    /* istanbul ignore if  */
    if (!task) {
        console.error('enumWindowsProc() task not exists');
        return true
    }
    switch (task.matchType) {
        case 'pid': {
            const buf = ref.alloc(win32Api.DTypes.HINSTANCE);
            const tid = user32.GetWindowThreadProcessId(hWnd, buf);

            /* istanbul ignore else */
            if (tid) {
                const pid = buf.readUInt32LE(0);

                if (pid && pid === task.matchValue) {
                    // console.log('got by pid:', pid, 'hWnd:' + ref.address(hWnd), ref.address(hWnd).toString(16));
                    task.hwndSet.add(hWnd);
                }
            }
            else {
                const lastErrMsg = get_last_err_msg();

                if (lastErrMsg) {
                    task.errMsg += ',' + lastErrMsg;
                }
            }
            break
        }
        case 'title':
            /* istanbul ignore else */
            if (task.matchValue) {
                const buf = Buffer.alloc(254);
                const len = user32.GetWindowTextW(hWnd, buf, 254);

                if (len) {
                    const name = buf.toString('ucs2');

                    // const visible = user32.IsWindowVisible(hWnd);
                    if (name.indexOf(task.matchValue) !== -1) {
                        const buf2 = Buffer.alloc(4);

                        user32.GetWindowThreadProcessId(hWnd, buf2);
                        task.pidSet.add(buf2.readUIntLE(0, 4));
                        task.hwndSet.add(hWnd);
                    }
                }
                else {
                    const lastErrMsg = get_last_err_msg();

                    if (lastErrMsg) {
                        task.errMsg += ',' + lastErrMsg;
                    }
                }
            }
            else { // all
                task.hwndSet.add(hWnd);
            }
            break
        case 'hwnd':
            if (+task.matchValue === ref.address(hWnd)) {
                task.hwndSet.add(hWnd);
            }
            break
        /* istanbul ignore next */
        default:
            return true
    }
    return true
});
function task_get_hwnds(task) {
    let t;

    return Promise.race([
        new Promise(resolve => {
            t = setTimeout(() => {
                console.info('task_get_hwnds timeout failed');
                resolve();
            }, 30000); // @HARDCOD
        }),
        _get_hwnds(task),
    ]).then((res) => {
        clearTimeout(t);
        task && taskConfig.task.delete(task.tno);
        if (res && Array.isArray(res) && res.length) {
            return res
        }
    }).catch(err => {
        clearTimeout(t);
        task && taskConfig.task.delete(task.tno);
        console.error(err);
    })
}
function _get_hwnds(task) {
    if (task.matchType === null) {
        if (typeof task.matchValue === 'string') {
            task.matchType = 'title';
        }
        else if (typeof task.matchValue === 'number') { // hwnd must be specified explicitly
            task.matchType = 'pid';
        }
    }
    switch (task.matchType) {
        case 'pid':
            if (!Number.isSafeInteger(task.matchValue)) {
                return Promise.resolve()
            }
            break
        case 'title':
            // void
            break
        case 'hwnd':
            // void
            break
        /* istanbul ignore next */
        default:
            throw new Error('task.matchType value invalid')
    }
    return get_task_hwnd(task)
        .then(res => {
        return res
    })
        .catch(err => {
        console.error(err);
        return
    })
}
function get_task_hwnd(task) {
    return new Promise((resolve, reject) => {
        if (!task) {
            resolve([]);
            return
        }
        user32.EnumWindows.async(enumWindowsProc, task.tno, (err) => {
            // console.log('get_task_hwnd size:', task.hwndSet.size);
            if (task.errMsg) {
                console.error(task.errMsg);
            }
            resolve(Array.from(task.hwndSet) || []);
        });
    })
}
function create_task() {
    const tno = ++taskConfig.tno;
    const task = {
        tno,
        matchType: null,
        matchValue: '',
        hwndSet: new Set(),
        pidSet: new Set(),
        errMsg: '',
    };

    taskConfig.task.set(tno, task);
    return task
}

const user32$1 = win32Api.U.load();
// kill process matched by pid|title, ExecRet.pids contains processed pid
function kill(opts) {
    const execRet = init_execret();

    if (opts.matchType === 'pid') {
        const pid = +opts.matchValue;

        if (pid <= 0) {
            execRet.err = 1;
            execRet.msg = 'pid value invalid';
            return Promise.resolve(execRet)
        }
        _kill(pid);
        execRet.pids.push(pid);
        return Promise.resolve(execRet)
    }
    else if (opts.matchType === 'title') {
        return new Promise(resolve => {
            const task = create_task();

            task.matchType = opts.matchType;
            task.matchValue = opts.matchValue;
            return task_get_hwnds(task).then(() => {
                if (task.pidSet.size) {
                    for (const pid of task.pidSet) {
                        if (_kill(pid)) {
                            execRet.pids.push(pid);
                        }
                    }
                }
                else {
                    execRet.msg = 'the pid list to be killed empty, none title matched';
                }
                resolve(execRet);
            })
        })
    }
    else {
        execRet.err = 1;
        execRet.msg = 'kill() matchType invalid';
        return Promise.resolve(execRet)
    }
}
function _kill(pid) {
    try {
        process.kill(pid, 0) && process.kill(pid);
        return true
    }
    catch (ex) {
        console.error(ex);
    }
    return false
}
function init_execret() {
    return {
        err: 0,
        msg: '',
        pids: [],
        hwnds: [],
    }
}
function parse_cli_opts(argv) {
    const opts = Object.assign({}, showFilterRulesDefaults);

    if (argv.pid) {
        opts.matchType = 'pid';
        opts.matchValue = +argv.pid;
    }
    else if (argv.title) {
        opts.matchType = 'title';
        opts.matchValue = argv.title;
    }
    else if (argv.hwnd) {
        opts.matchType = 'hwnd';
        opts.matchValue = argv.hwnd;
    }
    else {
        return
    }
    if (typeof argv.status !== 'undefined') {
        opts.nCmdShow = +argv.status;
        console.assert(typeof opts.nCmdShow === 'number', 'nCmdShow must typeof number');
    }
    return opts
}

/**
 * kill process by which the matched hWnd(s) (window) created
 * optins:
 * --pid={integer}
 * --title={string}
 * --hwnd={integer}
 */
const opts = parse_cli_opts(yargs.argv);

if (!opts) {
    console.error('argv empty. options: --title or --pid');
    process.exit(1);
}
else {
    kill(opts).then((execRet) => {
        console.info('process ret:', execRet);
        process.exit(execRet.err);
    });
}
