import { access, chmod, close, copyFile, mkdir, open, readdir, readFile, rmdir, stat, unlink, write, writeFile, } from 'fs'
import { basename, dirname, join, normalize, resolve as pathResolve, sep, } from 'path'
import { promisify } from 'util'
export const closeAsync = promisify(close)
export const chmodAsync = promisify(chmod)
export const copyFileAsync = promisify(copyFile)
export const mkdirAsync = promisify(mkdir)
export const openAsync = promisify(open)
export const readFileAsync = promisify(readFile)
export const readDirAsync = promisify(readdir)
export const rmdirAsync = promisify(rmdir)
export const unlinkAsync = promisify(unlink)
export const writeAsync = promisify(write)
export const writeFileAsync = promisify(writeFile)
export { basename, dirname, join, normalize, pathResolve, promisify, }
export { tmpdir } from 'os'
// support relative file ('./foo')
export function isPathAcessible(path) {
    return path
        ? new Promise(resolve => access(path, err => resolve(err ? false : true)))
        : Promise.resolve(false)
}
export function isDirExists(path) {
    return path ? isDirFileExists(path, 'DIR') : Promise.resolve(false)
}
export function isFileExists(path) {
    return path ? isDirFileExists(path, 'FILE') : Promise.resolve(false)
}
function isDirFileExists(path, type) {
    return path
        ? new Promise(resolve => {
            stat(path, (err, stats) => {
                err || !stats ? resolve(false) : resolve(type === 'DIR' ? stats.isDirectory() : stats.isFile())
            })
        })
        : Promise.resolve(false)
}
// create directories recursively
export async function createDir(path) {
    if (!path) {
        throw new Error('value of path param invalid')
    }
    else {
        path = normalize(path) // ! required for '.../.myca' under win32
        /* istanbul ignore else */
        if (!await isDirExists(path)) {
            await path.split(sep).reduce(async (parentDir, childDir) => {
                const curDir = pathResolve(await parentDir, childDir)

                await isPathAcessible(curDir) || await mkdirAsync(curDir, 0o755)
                return curDir
            }, Promise.resolve(sep))
        }
    }
}
export async function createFile(file, data, options) {
    const path = dirname(file)

    /* istanbul ignore next */
    if (!path) {
        throw new Error('path empty')
    }
    if (!await isDirExists(path)) {
        await createDir(path)
    }
    file = normalize(file)
    /* istanbul ignore else */
    if (!await isFileExists(file)) {
        const opts = options ? options : { mode: 0o640 }

        if (typeof data === 'object') {
            await writeFileAsync(file, JSON.stringify(data))
        }
        else {
            await writeFileAsync(file, data, opts)
        }
    }
}
/* istanbul ignore next */
export function logger(...args) {
    // tslint:disable-next-line
    console.log(args)
}
export function assertNever(x) {
    throw new Error('Assert Never Unexpected object: ' + x)
}
/**
 * Remove directory recursively
 * @see https://stackoverflow.com/a/42505874/3027390
 */
export async function rimraf(path) {
    if (!path) {
        return
    }
    await _rimraf(path)
    if (await isDirExists(path)) {
        await rmdirAsync(path)
    }
}
async function _rimraf(path) {
    if (!path) {
        return
    }
    if (await isPathAcessible(path)) {
        if (await isFileExists(path)) {
            await unlinkAsync(path)
            return
        }
        const entries = await readDirAsync(path)

        if (entries.length) {
            for (const entry of entries) {
                await _rimraf(join(path, entry))
            }
        }
        else {
            await rmdirAsync(path)
        }
    }
}
